from typing import Any, Sequence, TypeVar, Generic, Optional

from PySide6.QtCore import QAbstractItemModel, QModelIndex, QPersistentModelIndex
from PySide6.QtWidgets import QWidget

from .model_helpers import HasColumnData, HasParent, HasChildren

T = TypeVar("T")


# todo comment this
class TreeModel(QAbstractItemModel, Generic[T]):
    def __init__(self, parent: Optional[QWidget] = None):
        super().__init__(parent)
        self.root: Sequence[T] = []

    def get_item(self, index: QModelIndex | QPersistentModelIndex) -> object:
        if index.isValid():
            return index.internalPointer()

        return self.root

    def data(self, index: QModelIndex | QPersistentModelIndex, role: int = ...) -> Any:
        if not index.isValid():
            return None

        item = self.get_item(index)

        if isinstance(item, HasColumnData):
            return item.get_column_data(index.column(), role)
        else:
            return None

    def index(self, row: int, column: int, parent: QModelIndex | QPersistentModelIndex = ...) -> QModelIndex:
        if parent.isValid() and parent.column() != 0:
            return QModelIndex()

        parent_item = self.get_item(parent)

        if isinstance(parent_item, HasChildren):
            parent_list = parent_item.children
        elif isinstance(parent_item, Sequence):
            parent_list = parent_item
        else:
            raise TypeError(
                f"Expected parent of item to be Sequence or HasChildren, actually {parent_item.__class__}")

        if row > len(parent_list) or row < 0:
            return QModelIndex()

        return self.createIndex(row, column, parent_list[row])

    def parent(self, child: QModelIndex | QPersistentModelIndex = QModelIndex()) -> QModelIndex:
        if not child.isValid():
            return QModelIndex()

        child_item = self.get_item(child)

        # We don't need to handle the root items here because... they are the root and have no parents

        if isinstance(child_item, HasParent) and child_item.parent is not None:
            return self.createIndex(child_item.index_of_self, 0, child_item.parent)

        return QModelIndex()

    def rowCount(self, parent: QModelIndex | QPersistentModelIndex = QModelIndex()) -> int:
        if parent.isValid() and parent.column() > 0:
            return 0

        parent_item = self.get_item(parent)

        if isinstance(parent_item, HasChildren):
            return len(parent_item.children)
        elif isinstance(parent_item, Sequence):
            return len(parent_item)
        else:
            return 0

    def get_root(self) -> Sequence[T]:
        """
        Returns the list of root level items (i.e. items which have no parents) for the tree
        :return: List of root level items (i.e. items which have no parents) for the tree
        :return: List of root level items (i.e. items which have no parents) for the tree
        """
        return self.root



from PySide6.QtCore import QMimeDatabase
from PySide6.QtGui import QIcon
from PySide6.QtWidgets import QApplication, QStyle

_mime_database = QMimeDatabase()
_default_icon = None


def icon_for_filename(file_name: str) -> QIcon:
    """
    Gets the icon for a filename, based on its extension
    Parameters
    ----------
    file_name
        The name of the file
    Returns
    -------
    QIcon
        The icon for the file
    """
    global _default_icon

    for mime_type in _mime_database.mimeTypesForFileName(file_name):
        icon = QIcon.fromTheme(mime_type.iconName())

        # Return the appropriate icon if it's found
        if not icon.isNull():
            return icon

    # Cache the default icon, used when the icon for a file is unknown/not found
    if _default_icon is None:
        _default_icon = QApplication.style().standardIcon(QStyle.StandardPixmap.SP_FileIcon)

    return _default_icon



import logging
from abc import ABC, abstractmethod
from asyncio import Lock, Event

import httpx
from httpx import URL


class AuthenticationException(Exception):
    pass


# httpx does have an authentication flow mechanism that allows you to also make other requests but I don't know if it
# will behave the same way as this does. I also finished this before I found out that existed.

class SelfAuthenticatingWithHttpClient(ABC):
    _logger = logging.getLogger(__name__)

    def __init__(self, max_retires: int = 3, client: httpx.AsyncClient = httpx.AsyncClient()):
        self.authentication_lock = Lock()
        self.client = client
        self.authentication_in_progress: Event | None = None
        self.max_retries = max_retires

    async def reauthenticate(self) -> None:
        """
        Re-authenticates the client
        """

        # Acquire the lock as to prevent multiple threads from modifying self.authentication_in_progress.
        # Not that will ever happen using asyncio alone because python async is not actually multi-processed
        await self.authentication_lock.acquire()

        # If there's no authentication already in progress
        if self.authentication_in_progress is None:
            # Update authentication_in_progress and release the lock
            self.authentication_in_progress = Event()
            self.authentication_lock.release()

            # Start the authentication
            self._logger.info(f"Authenticating {self.__class__.__name__}")
            await self._authenticate()
            self._logger.info(f"Finished authenticating {self.__class__.__name__}")
            # Update the event to unblock other coroutines
            self.authentication_in_progress.set()

            # Delete authentication_in_progress to indicate that authentication is no longer in progress
            # This lock is probably not needed (due to not actually being multi-processed) but shouldn't hurt
            await self.authentication_lock.acquire()
            self.authentication_in_progress = None
            self.authentication_lock.release()
        # Authentication is already in progress
        else:
            # Get a reference to the event
            event = self.authentication_in_progress
            # Release the lock as we are not going to access it anymore
            self.authentication_lock.release()

            # Wait for the reauthentication to finish
            self._logger.info(f"Waiting for {self.__class__.__name__}")
            await event.wait()
            self._logger.info(f"Finished waiting for {self.__class__.__name__}")

    async def do_request_and_retry_if_unauthenticated(self, url: URL, method: str, **kwargs) -> httpx.Response:
        """
        Executes a http request or reauthenticate and retries if needed
        :param url: The url of the request
        :param method: The method the request should use (post, get, etc)
        :return:
        """
        retries = 0

        async def make_request():
            return await self.client.request(url=url, method=method, **kwargs)

        # Make the initial request
        response = await make_request()

        # Retry if canvas is trying to get us to reauthenticate
        while (await self.reauthenticate_if_needed(response)) and retries < self.max_retries:
            response = await make_request()
            retries += 1

        return response

    async def reauthenticate_if_needed(self, response: httpx.Response) -> bool:
        """
        Inspects a response and activates reauthentication if the response indicates we need to
        :param response: The response to inspect
        :return: True if reauthentication was activated, false if not
        """

        if self.detect_authentication_needed(response):
            await self.reauthenticate()
            return True

        return False

    @abstractmethod
    def detect_authentication_needed(self, response: httpx.Response) -> bool:
        ...

    @abstractmethod
    async def _authenticate(self) -> None:
        """
        Authenticates the client
        """
        ...



from bs4 import Tag
from httpx import URL

from qcanvas import db as db
from qcanvas.net.canvas import CanvasClient
from qcanvas.util.link_scanner.resource_scanner import ResourceScanner

canvas_resource_id_prefix = "canvas_file"


class CanvasFileScanner(ResourceScanner):
    _canvas_client: CanvasClient

    def __init__(self, canvas_client: CanvasClient):
        self._canvas_client = canvas_client

    def accepts_link(self, link: Tag) -> bool:
        if link.name not in ["a", "img"]:
            return False

        return "data-api-returntype" in link.attrs.keys() and link["data-api-returntype"] == "File"

    async def extract_resource(self, link: Tag, file_id: str) -> db.Resource:
        return db.convert_legacy_file(await self._canvas_client.get_file_from_endpoint(link["data-api-endpoint"]))

    def extract_id(self, link: Tag) -> str:
        # https://canvas.newcastle.edu.au/courses/27716/files/5975585/...
        # --------------------------------- Extract this part ^^^^^^^
        return URL(link["data-api-endpoint"]).path.rsplit('/', 2)[-1]

    async def download(self, resource):
        path = resource.download_location
        path.parent.mkdir(parents=True, exist_ok=True)

        with open(path, "wb") as file:
            async for progress in self._canvas_client.download_file(resource, file):
                yield progress

    @property
    def name(self) -> str:
        return canvas_resource_id_prefix



from typing import Sequence

from sqlalchemy.ext.asyncio import AsyncSession

import qcanvas.db as db
import qcanvas.queries as queries
from qcanvas.util.helpers import canvas_sanitiser


async def create_assignments(g_course: queries.Course, session: AsyncSession) -> Sequence[db.Assignment]:
    """
    Puts assignments for a course into the database
    """
    assignments = []

    for g_assignment in g_course.assignments_connection.nodes:
        assignment = await session.get(db.Assignment, g_assignment.q_id)

        if assignment is None:
            assignment = db.Assignment()
            assignment.id = g_assignment.q_id
            assignment.course_id = g_course.m_id
            session.add(assignment)
        # todo related to scanning old modulepages, but assignment content is already present in the query result so no more requests need to be made
        elif g_assignment.updated_at.replace(tzinfo=None) <= assignment.updated_at:
            continue

        assignment.name = canvas_sanitiser.remove_garbage_from_title(g_assignment.name)
        assignment.description = g_assignment.description
        assignment.created_at = g_assignment.created_at
        assignment.updated_at = g_assignment.updated_at
        assignment.due_at = g_assignment.due_at
        assignment.position = g_assignment.position

        assignments.append(assignment)

    return assignments


async def create_modules(g_course: queries.Course, session: AsyncSession):
    """
    Creates modules for a course and puts them in the database
    """
    for g_module in g_course.modules_connection.nodes:
        module = await session.get(db.Module, g_module.q_id)

        if module is None:
            module = db.Module()
            module.id = g_module.q_id
            module.course_id = g_course.m_id
            session.add(module)

        module.name = canvas_sanitiser.remove_garbage_from_title(g_module.name)


async def create_course(g_course: queries.Course, session: AsyncSession, term: db.Term):
    """
    Creates course entries in the database
    """
    course = await session.get(db.Course, g_course.m_id)
    if course is None:
        course = db.Course()
        course.id = g_course.m_id
        course.preferences = db.CoursePreferences()
        session.add(course)

    course.name = canvas_sanitiser.remove_garbage_from_title(g_course.name)
    course.term = term


async def create_term(g_course: queries.Course, session) -> db.Term:
    term = await session.get(db.Term, g_course.term.q_id)

    if term is None:
        term = db.convert_term(g_course.term)
        session.add(term)

    return term



import itertools
from abc import ABC, abstractmethod
from typing import Any


class ProgressSection:
    """
    Convenience class for reporting on a section of a larger task
    """

    def __init__(self, section_name: str, total_progress: int, reporter: "ProgressReporter"):
        self.total_progress = total_progress
        self.reporter = reporter
        self._counter = itertools.count(1)  # Start at 1 instead of 0
        self.reporter.section_started(section_name, total_progress)

    def increment_progress(self, *args):
        self.reporter.progress(next(self._counter), self.total_progress)


class ProgressReporter(ABC):
    """
    A progress reporter is passed to a function which then uses it to report the progress for whatever task it performs to another place.
    Only 1 section should be active at a time.
    """

    @abstractmethod
    def section_started(self, section_name: str, total_progress: int) -> None:
        """
        Signals that a new section of a task has started
        Parameters
        ----------
        section_name
            The name of the new task section
        """
        ...

    @abstractmethod
    def progress(self, current_progress: int, total_progress: int) -> None:
        """
        Updates the current progress of the current task
        Parameters
        ----------
        current_progress
            The current amount of progress, e.g. the number of bytes of a file that have been downloaded
        total_progress
            The total or final amount of progress, e.g. the size of a download
        """
        ...

    @abstractmethod
    def finished(self) -> None:
        """
        Signals that the task is finished and there is nothing left to do
        """
        ...

    @abstractmethod
    def errored(self, context: Any) -> None:
        """
        Signals that the task could not be completed
        Parameters
        ----------
        context
            Any information about why the task failed
        """
        ...

    def section(self, section_name: str, total_progress: int) -> ProgressSection:
        """
        Creates a ProgressSection for this reporter, for reporting on a section of a larger task
        Parameters
        ----------
        section_name
            The name of the section
        total_progress
            The total or final amount of progress, e.g. the size of a download
        Returns
        -------
        ProgressSection
            The created section
        """
        return ProgressSection(section_name, total_progress, self)


class _NoopReporter(ProgressReporter):

    def section_started(self, section_name: str, total_progress: int) -> None:
        pass

    def progress(self, current_progress: int, total: int) -> None:
        pass

    def finished(self) -> None:
        pass

    def errored(self, context: Any) -> None:
        pass


noop_reporter = _NoopReporter()
